using DicomObjects;
using DicomObjects.DicomUIDs;
using DicomObjects.Enums;
using DicomObjects.EventArguments;
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;

namespace PrintSCP
{
    /// <summary>
    /// Summary description for Form1.
    /// </summary>
    public class Form1 : Form
    {
        private TextBox LogText;
        
        private const string LOGURL = "C:\\dicom log files";
        private const string Manufacturer = "Medical Connections";
        private const string ManufacturerModelName = "Demo Printer SCP";
        private const string PrinterStatus = "NORMAL";
        private const string DriverSerialNumber = "serial no 1234";
        private const string PrinterName = "PrinterName";
        private const string DateFormat = "yyyyMMdd";
        private const string TimeFormat = "hhmmss";
        private const string ServerStartingMessage = "Server Started, Listen Port ";
        private const string ServerStartingErrorMessage = "ERROR! Failed to Start Server : ";
        private const string FontTimesNewRoman = "Times New Roman";
        private const string FontConsolas = "Consolas";
        private const string FontArial = "Arial";
        private const string AssociationRequestFromMessage = "Association Request from";
        private const string Association = "Association";
        private const string ClosedText = "Closed";
        private const string Portrait = "PORTRAIT";
        private const string StartingPrinting = "Starting Printing";
        private const string PrintingFatalError = "Fatal Error! ";
        private const string PrintingFinished = "Printing Finished";

        /// <summary>
        /// Required designer variable.
        /// </summary>
        private readonly System.ComponentModel.Container components = null;

        private DicomServer Server;
        public static DicomDataSet PrinterObject { get; set; }
        public class PrinterAssociation : DicomAssociation
        {
            public PrinterAssociation()
            {
                DataSets = new DicomDataSetCollection
                {
                    PrinterObject // enables N-GET of printer object                
                };
            }

            public DicomDataSetCollection DataSets { get; set; }
            public SessionDataSet FilmSession { get; set; }
            public MyPrintDoc MyPrinterDoc { get; set; }
        }

        public class SessionDataSet : DicomDataSet
        {
            public SessionDataSet()
            {
                FilmBoxes = new DicomDataSetCollection();
            }
            public DicomDataSetCollection FilmBoxes { get; set; }
        }

        public class MyPrintDoc : System.Drawing.Printing.PrintDocument
        {
            public MyPrintDoc(DicomDataSetCollection datasets)
            {
                PrintDataSets = datasets;
                UID = string.Empty;
                Status = 0xC000; // default value could be something more meaningful
            }

            public DicomDataSetCollection PrintDataSets { get; set; }
            public string UID { get; set; }
            public int Status { get; set; }
            public int CurrentPageNumber { get; set; }
            public int TotalPageNumber { get; set; }
        }

        public Form1()
        {
            //
            // Required for Windows Form Designer support
            //
            InitializeComponent();
        }

        /// <summary>
        /// Clean up any resources being used.
        /// </summary>
        protected override void Dispose(bool disposing)
        {
            if (disposing && components != null)
            {
                components.Dispose();
            }
            base.Dispose(disposing);
        }

        #region Windows Form Designer generated code
        /// <summary>
        /// Required method for Designer support - do not modify
        /// the contents of this method with the code editor.
        /// </summary>
        private void InitializeComponent()
        {
            this.LogText = new TextBox();
            Label label1 = new Label();
            this.SuspendLayout();
            // 
            // LogText
            // 
            this.LogText.BorderStyle = BorderStyle.FixedSingle;
            this.LogText.Location = new System.Drawing.Point(8, 32);
            this.LogText.Multiline = true;
            this.LogText.Name = "LogText";
            this.LogText.ScrollBars = ScrollBars.Vertical;
            this.LogText.Size = new System.Drawing.Size(560, 360);
            this.LogText.TabIndex = 0;
            // 
            // label1
            // 
            label1.AutoSize = true;
            label1.Location = new System.Drawing.Point(8, 8);
            label1.Name = "label1";
            label1.Size = new System.Drawing.Size(229, 13);
            label1.TabIndex = 1;
            label1.Text = "By default, DicomServer is listening on port 104";
            // 
            // Form1
            // 
            this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
            this.ClientSize = new System.Drawing.Size(576, 398);
            this.Controls.Add(label1);
            this.Controls.Add(this.LogText);
            this.Name = "Form1";
            this.Text = "Print SCP example";
            this.FormClosing += new FormClosingEventHandler(this.Form1_FormClosing);
            this.Load += new System.EventHandler(this.Form1_Load);
            this.ResumeLayout(false);
            this.PerformLayout();

        }
        #endregion

        /// <summary>
        /// The main entry point for the application.
        /// </summary>
        [STAThread]
        static void Main()
        {
            Application.Run(new Form1());
        }

        private void Log(string s)
        {
            AppendText(LogText, s);
        }

        delegate void SetTextCallback(TextBox box, string text);

        private void AppendText(TextBox box, string text)
        {
            // InvokeRequired required compares the thread ID of the
            // calling thread to the thread ID of the creating thread.
            // If these threads are different, it returns true.
            if (box.InvokeRequired)
            {
                SetTextCallback d = AppendText;
                Invoke(d, new object[] { box, text });
            }
            else
            {
                box.AppendText(text + "\r\n");
            }
        }

        private void DecodeFormat(string format, out int x, out int y)
        {
            int p, p1;
            p = format.IndexOf(@"\");
            p1 = format.IndexOf(",", p + 1);
            x = Convert.ToInt32(format.Substring(p + 1, p1 - p - 1));
            y = Convert.ToInt32(format.Substring(p1 + 1));
        }

        private void MakePrinterDataset()
        {
            //  Dummy values as placeholder
            DicomDataSet dataSet = new DicomDataSet
            {
                InstanceUID = WellKnownInstances.Printer
            };
            dataSet.Add(Keyword.SOPClassUID, SOPClasses.Printer);
            dataSet.Add(Keyword.Manufacturer, Manufacturer);
            dataSet.Add(Keyword.ManufacturerModelName, ManufacturerModelName);
            dataSet.Add(Keyword.PrinterStatus, PrinterStatus);
            dataSet.Add(Keyword.PrinterStatusInfo, PrinterStatus);
            dataSet.Add(Keyword.DeviceSerialNumber, DriverSerialNumber);
            dataSet.Add(Keyword.SoftwareVersions, DicomGlobal.AssemblyVersion);
            dataSet.Add(Keyword.PrinterName, PrinterName);
            dataSet.Add(Keyword.DateOfLastCalibration, DateTime.Now.Date.ToString(DateFormat));
            dataSet.Add(Keyword.TimeOfLastCalibration, DateTime.Now.TimeOfDay.ToString(TimeFormat));
            PrinterObject = dataSet;
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            int port = 104;
            Server = new DicomServer();
            try
            {
                DicomGlobal.LogToFile(LOGURL, 63);

                Server.Listen(port);
                Server.NewAssociationType = typeof(PrinterAssociation);
                Log(ServerStartingMessage + port.ToString());
                Server.AssociationRequest += Server_AssociationRequest;
                Server.AssociationClosed += Server_AssociationClosed;
                Server.NormalisedReceived += Server_NormalisedReceived;
                Server.VerifyReceived += (o, ee) => { ee.Status = 0; };
            }
            catch (DicomException ex)
            {
                Log(ServerStartingErrorMessage + ex.Message);
            }
            // Make a Global Printer Object
            MakePrinterDataset();

            initializeSupportedDisplayFomats();
        }

        class DisplayFormats
        {
            public DisplayFormats(int id, RectangleF area, Font fontType)
            {
                ID = id;
                Area = area;
                FontType = fontType;
            }
            public int ID;
            public RectangleF Area;
            public Font FontType;
        }

        void initializeSupportedDisplayFomats()
        {
            //  Defining SCP supported Annotation boxes
            DicomDataSet annotation1 = new DicomDataSet();
            annotation1.Add(Keyword.ReferencedSOPClassUID, SOPClasses.BasicAnnotationBox);
            annotation1.Add(Keyword.ReferencedSOPInstanceUID, DicomGlobal.NewUID());
            BasicAnnotationBoxes.Add(annotation1);

            DicomDataSet annotation2 = new DicomDataSet();
            annotation2.Add(Keyword.ReferencedSOPClassUID, SOPClasses.BasicAnnotationBox);
            annotation2.Add(Keyword.ReferencedSOPInstanceUID, DicomGlobal.NewUID());
            BasicAnnotationBoxes.Add(annotation2);

            //  Defining display formats for each AnnotationID
            SupportedDisplayFormats.Add(new DisplayFormats(1, new RectangleF(10, 10, 100, 100), new Font(FontTimesNewRoman, 10)));
            SupportedDisplayFormats.Add(new DisplayFormats(2, new RectangleF(150, 50, 100, 200), new Font(FontConsolas, 20)));
            SupportedDisplayFormats.Add(new DisplayFormats(3, new RectangleF(80, 150, 100, 200), new Font(FontArial, 8)));
        }

        private void Server_AssociationRequest(object sender, AssociationRequestArgs e)
        {
            Log($"{AssociationRequestFromMessage} {e.Association.CallingAET} At {e.Association.RemoteIP}");
            foreach (DicomContext ct in e.Contexts)
            {
                if (ct.AbstractSyntax == MetaSOPClasses.BasicGrayscalePrint)
                {
                    DicomDataSetCollection nullDataSetCollection = new DicomDataSetCollection();
                    var myPrinterDoc = new MyPrintDoc(nullDataSetCollection)
                    {
                        CurrentPageNumber = 0,
                        TotalPageNumber = 0
                    };
                    myPrinterDoc.PrintPage += MyPrinterDoc_PrintPage;
                    myPrinterDoc.EndPrint += MyPrinterDoc_EndPrint;
                    e.Association.Tag = myPrinterDoc;
                }
            }
        }

        private void Server_AssociationClosed(object sender, AssociationClosedArgs e)
        {
            Log($"{Association} {e.Association.AssociationNumber} {ClosedText}");
        }

        readonly DicomDataSetCollection BasicAnnotationBoxes = new DicomDataSetCollection();
        DicomLabelCollection AnnotationLabels = null;
        readonly List<DisplayFormats> SupportedDisplayFormats = new List<DisplayFormats>();


        private void Server_NormalisedReceived(object sender, NormalisedReceivedArgs e)
        {
            PrinterAssociation assoc = (PrinterAssociation)e.RequestAssociation;
            assoc.MyPrinterDoc = e.RequestAssociation.Tag as MyPrintDoc;

            DicomDataSetCollection DataSets = assoc.DataSets;
            DicomDataSet filmBox = null;
            DicomDataSetCollection ImageBoxes;
            DicomDataSet imageBox;
            DicomDataSet command = e.Command;

            int operation = Convert.ToInt32(command[Keyword.CommandField].Value);
            string requestedSOPClass = command[Keyword.RequestedSOPClassUID].Value + "";
            string requestedSOPInstanceUID = command[Keyword.RequestedSOPInstanceUID].Value + "";
            string affectedSOPClass = command[Keyword.AffectedSOPClassUID].Value + "";
            string affectedSOPInstanceUID = command[Keyword.AffectedSOPInstanceUID].Value + "";

            switch (operation)
            {
                case 0x110: // N-GET
                    filmBox = DataSets[requestedSOPInstanceUID];
                    e.ResultDataSet = filmBox;
                    e.Status = 0;
                    break;

                case 0x140: // N-CREATE
                    filmBox = e.Request.Clone() as DicomDataSet; // Set Default Values					
                    filmBox.Add(Keyword.SOPClassUID, affectedSOPClass);

                    if (affectedSOPInstanceUID == "")
                        affectedSOPInstanceUID = DicomGlobal.NewUID();
                    filmBox.Add(Keyword.SOPInstanceUID, affectedSOPInstanceUID);

                    // FilmSession
                    if (affectedSOPClass == SOPClasses.BasicFilmSession)
                    {
                        assoc.FilmSession = new SessionDataSet();
                        foreach (DicomAttribute attr in filmBox)
                        {
                            assoc.FilmSession.Add(attr.Group, attr.Element, attr.Value);
                        }
                        assoc.FilmSession.ValidationTypes = ValidationTypes.None;
                        DataSets.Add(assoc.FilmSession);
                    }

                    // Specific to FilmBox
                    if (affectedSOPClass == SOPClasses.BasicFilmBox)
                    {
                        DecodeFormat(filmBox[Keyword.ImageDisplayFormat].Value.ToString(), out int cols, out int rows);
                        ImageBoxes = new DicomDataSetCollection();

                        // e.PresentationContextId is only available in DicomObjects Version 5.7 and onwards
                        bool color = e.RequestAssociation.AgreedContexts[e.PresentationContextId].AbstractSyntax == MetaSOPClasses.BasicColorPrint.ToString();

                        for (int i = 1; i <= rows * cols; i++)
                        {
                            imageBox = new DicomDataSet
                            {
                                InstanceUID = DicomGlobal.NewUID()
                            };
                            if (!color)
                                imageBox.Add(Keyword.ReferencedSOPClassUID, SOPClasses.BasicGrayscaleImageBox);
                            else
                                imageBox.Add(Keyword.ReferencedSOPClassUID, SOPClasses.BasicColorImageBox);

                            imageBox.Add(Keyword.ReferencedSOPInstanceUID, imageBox.InstanceUID);
                            DataSets.Add(imageBox);

                            ImageBoxes.Add(imageBox);
                        }

                        filmBox.Add(Keyword.ReferencedImageBoxSequence, ImageBoxes); // Referenced Image Box Sequence

                        //  Check if SOP class has been proposed, only then add it.
                        foreach (var agreedContext in e.RequestAssociation.AgreedContexts)
                            if (agreedContext.AbstractSyntax == SOPClasses.BasicAnnotationBox.ToString())
                                filmBox.Add(Keyword.ReferencedBasicAnnotationBoxSequence, BasicAnnotationBoxes);

                        // Also Add FilmBox to current session
                        assoc.FilmSession.FilmBoxes.Add(filmBox);

                        DataSets.Add(filmBox);
                    }

                    e.ResultDataSet = filmBox;
                    e.Status = 0;
                    break;

                case 0x120: // N-SET
                    foreach (DicomDataSet ba in BasicAnnotationBoxes)
                        if (ba[Keyword.ReferencedSOPInstanceUID].Value.ToString() == requestedSOPInstanceUID)
                            filmBox = ba;
                    if (filmBox == null)
                        filmBox = DataSets[requestedSOPInstanceUID];
                    else
                        GenerateAnnotationLabels(e.Request);
                    foreach (DicomAttribute a in e.Request)
                    {
                        filmBox.Add(a.Group, a.Element, a.Value);
                    }
                    e.Status = 0;
                    break;

                case 0x130: // N-ACTION					

                    if (requestedSOPClass == SOPClasses.BasicFilmBox)       //  BasicFilmBox  is provided in the request
                    {
                        //  Increment pages
                        assoc.MyPrinterDoc.CurrentPageNumber += 1;

                        e.Status = PrintFilmBox(assoc, requestedSOPInstanceUID, DataSets);
                    }
                    else //  We assume that BasicFilmSession is provided instead
                    {
                        int status = 0xC000;

                        //  Iterate and Print each FilmBox
                        foreach (DicomDataSet filmbox in assoc.FilmSession.FilmBoxes)
                        {
                            //  Set total pages since we have the complete session details
                            assoc.MyPrinterDoc.TotalPageNumber = assoc.FilmSession.FilmBoxes.Count;
                            //  And print 1 at a time
                            assoc.MyPrinterDoc.CurrentPageNumber = 1;

                            status = PrintFilmBox(assoc, filmbox.InstanceUID, DataSets);
                        }
                        e.Status = status;
                    }
                    break;

                case 0x150: //N-Delete
                    DataSets.Remove(requestedSOPInstanceUID);

                    if (requestedSOPClass == SOPClasses.BasicFilmBox)
                        assoc.FilmSession.FilmBoxes.Remove(requestedSOPInstanceUID);

                    if (requestedSOPClass == SOPClasses.BasicFilmSession)
                        assoc.FilmSession = null;

                    e.Status = 0;
                    break;
            }
        }

        private int PrintFilmBox(PrinterAssociation assoc, string requestedSOPInstanceUID, DicomDataSetCollection DataSets)
        {
            assoc.MyPrinterDoc.UID = requestedSOPInstanceUID;
            assoc.MyPrinterDoc.PrintDataSets = DataSets;

            DicomDataSet filmbox = DataSets[requestedSOPInstanceUID];
            string orientation = filmbox[Keyword.FilmOrientation].Value.ToString();
            if (orientation == Portrait)
            {
                assoc.MyPrinterDoc.DefaultPageSettings.Landscape = false;
            }
            else
            {
                assoc.MyPrinterDoc.DefaultPageSettings.Landscape = true;
            }

            //  Print Document
            assoc.MyPrinterDoc.Print();

            //  return Status
            return assoc.MyPrinterDoc.Status;
        }

        private void GenerateAnnotationLabels(DicomDataSet requestDS)
        {
            DisplayFormats chosenDisplayFormat = null;
            foreach (var df in SupportedDisplayFormats)
            {
                if (df.ID == int.Parse(requestDS[Keyword.AnnotationPosition].Value.ToString()))
                    chosenDisplayFormat = df;

                if (chosenDisplayFormat != null)
                {
                    DicomLabel l = new DicomLabel()
                    {
                        Text = requestDS[Keyword.TextString].Value.ToString(),
                        Area = chosenDisplayFormat.Area,
                        Font = chosenDisplayFormat.FontType
                    };

                    if (AnnotationLabels == null)
                        AnnotationLabels = new DicomLabelCollection();
                    AnnotationLabels.Add(l);
                    chosenDisplayFormat = null;
                }
            }
        }

        private void MyPrinterDoc_PrintPage(object sender, System.Drawing.Printing.PrintPageEventArgs e)
        {
            Log(StartingPrinting);

            MyPrintDoc mp = (MyPrintDoc)sender;

            DicomDataSetCollection datasets = mp.PrintDataSets;
            DicomDataSet filmbox = datasets[mp.UID];
            DicomDataSetCollection imageBoxes = filmbox[Keyword.ReferencedImageBoxSequence].Value as DicomDataSetCollection;


            DecodeFormat(filmbox[Keyword.ImageDisplayFormat].Value.ToString(), out int columns, out int rows);

            int margin = 5;
            int width, height;
            int paperWidth, paperHeight;

            if (!mp.DefaultPageSettings.Landscape)
            {
                paperWidth = mp.PrinterSettings.PaperSizes[0].Width;
                paperHeight = mp.PrinterSettings.PaperSizes[0].Height;
            }
            else
            {
                paperWidth = mp.PrinterSettings.PaperSizes[0].Height;
                paperHeight = mp.PrinterSettings.PaperSizes[0].Width;
            }

            width = ((paperWidth - margin) / columns) - margin;
            height = ((paperHeight - margin) / rows) - margin;

            Point LeftTop = new Point();
            Font textFont = new Font("Arial", 8);
            string pageNumber = "Page " + mp.CurrentPageNumber.ToString();
            e.Graphics.DrawString(pageNumber, textFont, Brushes.Blue, 4, 4);

            for (int row = 0; row < rows; row++)
            {
                for (int column = 0; column < columns; column++)
                {
                    DicomDataSet ds2 = datasets[imageBoxes[row * columns + column][Keyword.ReferencedSOPInstanceUID].Value.ToString()];
                    DicomAttribute imageAtt;

                    // if Mono Image
                    if (ds2[Keyword.BasicGrayscaleImageSequence].Exists)
                    {
                        imageAtt = ds2[Keyword.BasicGrayscaleImageSequence];
                    }
                    // if not mono image, then check for a color image
                    else
                    {
                        imageAtt = ds2[Keyword.BasicColorImageSequence];
                    }

                    if (imageAtt.Exists) // has real image then print
                    {
                        DicomDataSetCollection dss2 = (DicomDataSetCollection)imageAtt.Value;
                        // Get the DicomImage to print
                        DicomImage dcmImage = new DicomImage(dss2[0]);
                        // Set up the Print Area						 
                        int min;
                        int offSetX = 0;
                        int offSetY = 0;
                        if (width > height)
                        {
                            min = height;
                            offSetX = (paperWidth - min * columns - (columns + 1) * margin) / 2;
                        }
                        else
                        {
                            min = width;
                            offSetY = (paperHeight - min * rows - (rows + 1) * margin) / 2;
                        }

                        LeftTop.X = column * (min + margin) + margin + offSetX;
                        LeftTop.Y = row * (min + margin) + margin + offSetY;

                        Rectangle PrintArea = new Rectangle(LeftTop.X, LeftTop.Y, min, min);

                        //  Add Annotation Labels if any
                        if (AnnotationLabels != null && AnnotationLabels.Count > 0)
                        {
                            dcmImage.Labels = AnnotationLabels;
                            AnnotationLabels = null;
                        }

                        // Get Bitmap Image by DicomImage.Bitmap() method
                        Image printImage = dcmImage.Bitmap();

                        try
                        {
                            // Print Image here
                            e.Graphics.DrawImage(printImage, PrintArea);
                            mp.Status = 0;
                        }
                        catch (Exception ex)
                        {
                            Log(PrintingFatalError + ex.Message);
                        }
                    }
                }
            }
        }

        private void MyPrinterDoc_EndPrint(object sender, System.Drawing.Printing.PrintEventArgs e)
        {
            Log(PrintingFinished);
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            Server.UnlistenAll();
        }

    }
}
